# -*- coding: UTF-8 -*-

from collections import OrderedDict
import sys


class Query(object):
    """ MySQL запрос """

    replaceSymbol = '%s'  # символ для подстановки значений в итоговый запрос

    def __init__(self, table, cacheFields, cacheValues, additionalKeyPart="", executeMany=True):
        """ MySQL запрос
        Args:
            table (str): Таблица
            cacheFields (list[str]): Названия полей/поля для взаимодействия
            cacheValues (list[str|int|decimal|_Select]): Значения полей
            additionalKeyPart (str): Дополнительная часть ключа кэша
            executeMany (bool): Выполнять запросы пачками или нет
        """
        self.table = table
        self.executeMany = executeMany

        # len(cacheFields) != len(cacheValues) нужно на случай если есть подзапрос, возвращающий значения нескольких колонок. Такое бывает
        if self.executeMany or len(cacheFields) != len(cacheValues):
            self.cacheKey = "%s@%s" % (self.table, repr(cacheFields))
        else:
            self.cacheKey = "%s@" % self.table
            for i in range(len(cacheFields)):
                self.cacheKey += list(cacheFields)[i]
                if isinstance(list(cacheValues)[i], _Select) and list(cacheValues)[i].executeMany:
                    self.cacheKey += "(%s)" % list(cacheValues)[i].cacheKey
        self.cacheKey += additionalKeyPart

    @staticmethod
    def makeList(value):
        """ Сформировать список из переданного значения
        Args:
            value (any): Значение
        Returns:
            list: Переданное значение в виде списка
        """
        if type(value) is list:
            return value
        elif type(value) is tuple:
            return list(value)
        else:
            return [value]

    def formWhereClause(self, context, fields, values):
        """ Сформировать общий вид условия запроса
        Args:
            context (Context): Контекст выполнения запросов
            fields (list[str]): Названия полей/поля для проверки условия
            values (list[str|int|decimal|_Select]): Параметры для подстановки в запрос
        Returns:
            str: Текст условия с местами для подстановки значений
        """
        where = "WHERE "
        size = len(fields)
        if self.executeMany and size == 1:
            where += "`%s` IN %s" % (fields[0], self.replaceSymbol)
        else:
            if size == len(values):
                for i in range(size):
                    where += " `%s` " % fields[i]
                    # если запрос одиночный и в нём есть Select, возвращающий одно значение, подставляем его в текст
                    if isinstance(values[i], _Select):
                        where += "IN (%s)" % values[i].prepareStatement(context)
                    else:
                        where += "= %s" % self.replaceSymbol
                    if i < size - 1:
                        where += " AND"
            else:
                where += (' AND '.join(['`%s` = ?'] * len(fields)) % tuple(fields)).replace('?', self.replaceSymbol)

        return where

    def convertValues(self, context, values, foreignCash=None):
        """ Сконвертировать значения параметров в подходящий вид
        Args:
            context (Context): Контекст выполнения запросов
            values (list[any]): Значения параметров
            foreignCash (str|None): Внешний ключ кэшироанния
        Returns:
            list|None: Преобразованные параметры
        """
        result = []
        for value in values:
            if sys.version_info.major == 2 and isinstance(value, unicode):
                result.append(value.encode("utf8"))
            elif isinstance(value, _Select):
                if self.executeMany or not value.executeMany:
                    selRaw = value.execute(context, foreignCash)
                    if type(selRaw) in (list, tuple) and len(selRaw) > 0:
                        if value.executeMany:
                            result.extend(selRaw[0])
                        else:
                            # если в условии запроса один параметр, то добаляем результаты Select в общий список
                            if len(values) == 1:
                                for row in selRaw:
                                    result.extend(row)
                            # если НЕ один, то добаляем все результаты как один элемент
                            else:
                                sel = []
                                for row in selRaw:
                                    sel.extend(row)
                                result.append(sel)
                    else:
                        return None
                else:
                    # добавляем параметры Select к параметрам основного запроса
                    result.extend(value.convertValues(context, value.whereValues))
            elif isinstance(value, dict):
                result.append(self.convertValues(context, value.values()))
            elif type(value) in (list, tuple):
                result.append(value)
            else:
                result.append(str(value))
        return result

    @staticmethod
    def execSql(context, queryText, values=None):
        """ Выполнить MySQL одиночный/групповой запрос
        Args:
            context (Context): Контекст выполнения запросов
            queryText (str): Текст MySQL запроса
            values (list[str|int|decimal|list[str|int|decimal]]|None): Параметры для подстановки в запрос
        """
        try:
            if values is not None:
                if len(values) > 0:
                    values = [list(v) if isinstance(v, {}.values().__class__) else v for v in values]
                    if type(values[0]) in (list, tuple):
                        context.cursor.executemany(queryText, values)
                    else:
                        context.cursor.execute(queryText, values)
                else:
                    context.log.warn("НЕ выполняется запрос с текстом \"%s\" из-за пустого списка параметров" % queryText)
            else:
                context.cursor.execute(queryText)
        except Exception as e:
            current_cursor = context.cursor
            if sys.version_info.major == 3 and sys.version_info.minor >= 7:
                executed_query = current_cursor._executed
            else:
                executed_query = current_cursor._last_executed

            context.log.error(u'Запрос с исходным текстом: \n"%s"\nи параметрами: \n"%s"\nпосле преобразования: \n"%s"\n'
                    u'при выполнении вызвал ошибку "%s"' % (queryText, values, executed_query, e))
            #Не проверяем на Duplicate entry, всё равно везде INSERT IGNORE
            raise e


class _Insert(Query):
    """ MySQL запрос 'Insert' """

    cache = {}  # сохранённые тексты SQL запросов
    cacheValues = {}  # сохранённые значения для выполнения запросов
    cacheOddColumns = {}  # кэш со столбцами, которые пропускаем (не пытаемся грузить в БД)
    cacheOwnerStoreIdPresented = {}  # если поле ownerStoreId предоставлено в таблице,
    # при insert его нужно установить в NULL, в ином случае при обогащении данных из All
    # запись для магазина не будет помечена, что она теперь из магазина и ее нужно выгружать"""

    def __init__(self, table, values, saveLastId=False, onDuplicateKeyUpdate=False):
        """ MySQL запрос 'Insert'
        Args:
            table (str): Таблица для вставки
            values (dict[str, str|int|decimal]): Значения для вставки. Ключами словаря должны быть названия колонок
            saveLastId (bool): Сохранить Id последней добавленной записи
            onDuplicateKeyUpdate (bool): Обновить значения при совпадении ключа
        """
        super(_Insert, self).__init__(table, values.keys(), values.values(), str(onDuplicateKeyUpdate), not saveLastId)
        self.values = values
        self.id = None
        self.saveLastId = saveLastId
        self.onDuplicateKeyUpdate = onDuplicateKeyUpdate

    def __getitem__(self, x):
        return self.values[x]

    def __setitem__(self, x, v):
        self.values[x] = v

    def __str__(self):
        return '[INSERT table=%s, values=%s, saveLastId=%s, onDuplicateKeyUpdate=%s]' % (self.table, self.values, self.saveLastId, self.onDuplicateKeyUpdate)

    def lastId(self):
        """ Получить Id последней добавленной записи
        Returns:
            int: Id последней добавленной записи
        """
        return self.id

    def filterColumns(self, context):
        """ Проверить наличие в базе данных запрошенных колонок и выбросить лишние
        Args:
            context (Context): Контекст выполнения запросов
        """
        if self.table not in _Insert.cacheOddColumns:
            _Insert.cacheOddColumns[self.table] = []
            realColumns = set()
            self.execSql(context, 'SHOW COLUMNS FROM `%s`' % self.table)
            for value in context.cursor:
                realColumns.add(value[0].lower())
            _Insert.cacheOwnerStoreIdPresented[self.table] = "ownerstoreid" in realColumns
            for k in list(self.values):
                if not k.lower() in realColumns:
                    _Insert.cacheOddColumns[self.table].append(k)
                    self.values.pop(k)
            if len(_Insert.cacheOddColumns[self.table]) > 0:
                context.log.warn('Не загружаем поля: %s в таблицу: %s' % (_Insert.cacheOddColumns[self.table], self.table))
        else:
            for k in _Insert.cacheOddColumns[self.table]:
                self.values.pop(k)

    def prepareStatement(self, context):
        """ Подготовить текст SQL запроса
        Args:
            context (Context): Контекст выполнения запросов
        Returns:
            str: Текст SQL запроса с местами для подстановки значений
        """
        if self.cacheKey in _Insert.cache:
            return _Insert.cache[self.cacheKey]
        fields = self.values.keys()
        sql = 'INSERT IGNORE INTO `%s` (`%s`) VALUES (%s)' % (context.tableName(self.table), '`,`'.join(fields),
                ','.join([self.replaceSymbol]*len(fields)))

        if self.onDuplicateKeyUpdate:
            onDuplicateKeyClause = ""
            for field in fields:
                if onDuplicateKeyClause:
                    onDuplicateKeyClause += ', '
                onDuplicateKeyClause += '`{0}` = VALUES(`{0}`)'.format(field)
            sql += ' ON DUPLICATE KEY UPDATE %s' % (onDuplicateKeyClause)

        _Insert.cache[self.cacheKey] = sql
        return sql

    def execute(self, context):
        """ Выполнить SQL запрос, либо добавить его значения в кэш для группового выполнения
        Args:
            context (Context): Контекст выполнения запросов
        """
        self.filterColumns(context)
        values = self.values.values()
        sql = self.prepareStatement(context)
        if self.executeMany:
            _Insert.cacheValues.setdefault(self.cacheKey, []).append(values)
        else:
            self.execSql(context, sql, values)
            if self.saveLastId:
                self.execSql(context, 'SELECT LAST_INSERT_ID()')
                for value in context.cursor:
                    self.id = value[0]

    @staticmethod
    def execMany(context):
        """ Выполнить SQL запросы, из кэша
        Args:
            context (Context): Контекст выполнения запросов
        """
        for key in list(_Insert.cacheValues):
            Query.execSql(context, _Insert.cache[key], _Insert.cacheValues.pop(key))


class _Update(Query):
    """ MySQL запрос 'Update' """
    cache = {}  # сохранённые тексты SQL запросов
    cacheValues = {}  # сохранённые значения для выполнения запросов

    def __init__(self, table, fields, values, whereField, whereValue):
        """ MySQL запрос 'Update'
        Args:
            table (str): Таблица для обновления данных
            fields (list[str]): Названия колонок для обновления данных
            values (list[str|int|decimal]): Новые значения
            whereField (str): Поле для проверки соответствия строки
            whereValue (str|_Select): Значение поля для проверки соответствия строки
        """
        self.fields = self.makeList(fields)
        self.values = self.makeList(values)
        self.whereFields = self.makeList(whereField)
        self.whereValues = self.makeList(whereValue)
        Query.__init__(self, table, self.fields + self.whereFields, self.values + self.whereValues)

    def __str__(self):
        return '[UPDATE table=%s, fields=%s, values=%s, whereField=%s, whereValue=%s]' % (self.table, self.fields, self.values, self.whereFields, self.whereValues)

    def prepareStatement(self, context):
        """ Подготовить текст SQL запроса
        Args:
            context (Context): Контекст выполнения запросов
        Returns:
            str: Текст SQL запроса с местами для подстановки значений
        """
        if self.cacheKey in _Update.cache:
            return _Update.cache[self.cacheKey]
        tableName = context.tableName(self.table)
        setClause = (', '.join(["`%s` = ?"] * len(self.fields)) % tuple(self.fields)).replace('?', self.replaceSymbol)
        sql = 'UPDATE `%s` SET %s %s' % (tableName, setClause, self.formWhereClause(context, self.whereFields, self.whereValues))

        _Update.cache[self.cacheKey] = sql
        return sql

    def execute(self, context):
        """ Выполнить SQL запрос
        Args:
            context (Context): Контекст выполнения запросов
        """

        whereValues = self.convertValues(context, self.whereValues)
        if whereValues is not None:
            sql = self.prepareStatement(context)
            values = self.values
            values.append(whereValues)
            if self.executeMany:
                _Update.cacheValues.setdefault(self.cacheKey, []).append(values)
            else:
                self.execSql(context, sql, values)

    @staticmethod
    def execMany(context):
        """ Выполнить SQL запросы, из кэша
        Args:
            context (Context): Контекст выполнения запросов
        """
        for key in list(_Update.cacheValues):
            Query.execSql(context, _Update.cache[key], _Update.cacheValues.pop(key))


class _Select(Query):
    """ MySQL запрос 'Select' """

    cache = {}  # сохранённые тексты SQL запросов
    cacheValues = {}  # сохранённые значения для выполнения запросов по внешним ключам кэша
    cacheUnite = {}  # сохранённые параметры объединения данных в один запрос

    def __init__(self, table, fields, whereFields, whereValues, arbitraryCondition=''):
        """ MySQL запрос 'Select'
        Args:
            table (str): Таблица для выборки данных
            fields (list[str]): Названия колонок для выборки
            whereFields (list[str]): Колонки для проверки соответствия строки
            whereValues (list[str|int|decimal]|_Select): Значения колонок для проверки соответствия строки
            arbitraryCondition (str): Произвольное условие SQL
        """
        self.fields = self.makeList(fields)
        self.whereFields = self.makeList(whereFields)
        self.whereValues = self.makeList(whereValues)
        self.result = False
        self.arbitraryCondition = arbitraryCondition
        Query.__init__(self, table, self.whereFields, self.whereValues, "@%s@%s" % (repr(self.fields), self.arbitraryCondition), len(self.fields) == 1)

    def __str__(self):
        return '[SELECT table=%s, fields=%s, whereFields=%s, whereValues=%s, arbitraryCondition=%s]' % (self.table, self.fields, self.whereFields, self.whereValues, self.arbitraryCondition)

    def prepareStatement(self, context):
        """ Подготовить текст SQL запроса
        Args:
            context (Context): Контекст выполнения запросов
        Returns:
            str: Текст SQL запроса с местами для подстановки значений
        """
        if self.cacheKey in _Select.cache:
            return _Select.cache[self.cacheKey]
        tableName = context.tableName(self.table)
        whereClause = self.formWhereClause(context, self.whereFields, self.whereValues)
        sql = 'SELECT `%s` FROM `%s` %s %s' % ('`,`'.join(self.fields), tableName, whereClause, self.arbitraryCondition)
        _Select.cache[self.cacheKey] = sql
        _Select.cacheUnite[self.cacheKey] = len(self.whereFields) == 1
        return sql

    def execute(self, context, foreignCash=None):
        """ Выполнить SQL запрос, либо добавить его значения в кэш для группового выполнения
        Args:
            context (Context): Контекст выполнения запросов
            foreignCash (str|None): Внешний ключ для кэшироанния
        Returns:
            list: Список результатов запроса
        """
        if not self.result:
            whereValues = self.convertValues(context, self.whereValues)
            if whereValues is not None:
                sql = self.prepareStatement(context)
                unite = _Select.cacheUnite[self.cacheKey]
                if foreignCash is not None:
                    cache = _Select.cacheValues.setdefault(foreignCash, {}).setdefault(self.cacheKey, [])
                    if unite:
                        cache.extend(whereValues)
                    else:
                        cache.append(whereValues)
                    self.result = True
                else:
                    if unite:
                        whereValues = [[whereValues]]
                    self.execSql(context, sql, whereValues)
                    self.result = context.cursor.fetchall()
        return self.result

    @staticmethod
    def execMany(context, foreignCash):
        """ Выполнить SQL запрос, либо добавить его значения в кэш для группового выполнения
        Args:
            context (Context): Контекст выполнения запросов
            foreignCash (str): Внешний ключ кэшироанния
        Returns:
            list: Список результатов запроса
        """
        result = []
        if foreignCash in list(_Select.cacheValues):
            for key, values in _Select.cacheValues.pop(foreignCash).items():
                if _Select.cacheUnite[key]:
                    values = [[values]]
                Query.execSql(context, _Select.cache[key], values)
                for item in context.cursor:
                    result.extend(item)
        return result


class _Delete(Query):
    """ MySQL запрос 'Delete' """

    cache = {}  # сохранённые тексты SQL запросов
    cacheValues = OrderedDict()  # сохранённые значения для выполнения запросов. Порядок важен при наличии Select'ов
    cacheUnite = {}  # сохранённые параметры объединения данных в один запрос

    def __init__(self, table, whereFields, whereValues, notExecMany=False):
        """ MySQL запрос 'Delete'
        Args:
            table (str): Таблица для удаления данных
            whereFields (list[str]): Поля для проверки соответствия строки
            whereValues (list[str|int|decimal]|_Select): Значения полей для проверки соответствия строки
            notExecMany (bool): Не применять групповое выполнение для этого запроса
        """
        self.whereFields = self.makeList(whereFields)
        self.whereValues = self.makeList(whereValues)
        self.notExecMany = notExecMany
        Query.__init__(self, table, self.whereFields, self.whereValues, executeMany=not notExecMany)

    def __str__(self):
        return '[DELETE table=%s, whereFields=%s, whereValues=%s]' % (self.table, self.whereFields, self.whereValues)

    def prepareStatement(self, context):
        """ Подготовить текст SQL запроса
        Args:
            context (Context): Контекст выполнения запросов
        Returns:
            str: Текст SQL запроса с местами для подстановки значений
        """
        if self.cacheKey in _Delete.cache:
            return _Delete.cache[self.cacheKey]
        tableName = context.tableName(self.table)
        sql = 'DELETE FROM `%s` %s' % (tableName, self.formWhereClause(context, self.whereFields, self.whereValues))
        _Delete.cache[self.cacheKey] = sql
        _Delete.cacheUnite[self.cacheKey] = len(self.whereFields) == 1
        return sql

    def execute(self, context):
        """ Выполнить SQL запрос, либо добавить его значения в кэш для группового выполнения
        Args:
            context (Context): Контекст выполнения запросов
        """
        if self.table not in context.emptyTables:
            sql = self.prepareStatement(context)
            unite = _Delete.cacheUnite[self.cacheKey]
            if self.executeMany:
                whereValues = self.convertValues(context, self.whereValues, self.cacheKey)
                cache = _Delete.cacheValues.setdefault(self.cacheKey, [])
                if whereValues is not None:
                    if unite:
                        cache.extend(whereValues)
                    else:
                        cache.append(whereValues)
            else:
                whereValues = self.convertValues(context, self.whereValues)
                if whereValues is not None:
                    if unite:
                        whereValues = [[whereValues]]
                    self.execSql(context, sql, whereValues)

    @staticmethod
    def execMany(context):
        """ Выполнить SQL запросы, из кэша
        Args:
            context (Context): Контекст выполнения запросов
        """
        for key in list(_Delete.cacheValues):
            values = _Delete.cacheValues.pop(key)
            if _Delete.cacheUnite[key]:
                values.extend(_Select.execMany(context, key))
                if len(values) == 0:
                    continue
                values = [[values]]
            Query.execSql(context, _Delete.cache[key], values)


class _DeleteAll(Query):
    """ Очистка таблицы """

    def __init__(self, table):
        """ Очистка таблицы
        Args:
            table (str): Таблица для очистки
        """
        super(_DeleteAll, self).__init__(table, [], [])

    def __str__(self):
        return "[DELETEALL table=%s]" % self.table

    def execute(self, context):
        """ Выполнить SQL запрос
        Args:
            context (Context): Контекст выполнения запросов
        """
        self.execSql(context, "TRUNCATE `%s`" % context.tableName(self.table))
        context.emptyTables[self.table] = True

    def execMany(self, context):
        """ Выполнить SQL запрос
        Args:
            context (Context): Контекст выполнения запросов
        """
        self.execute(context)


class _DeleteByDate(_Delete):
    """ Удаление переоценок по дате """

    def __init__(self, table, whereFields, whereValues, notExecMany=False):
        """ Удаление переоценок по дате
        Args:
            table (str): Таблица для удаления данных
            whereFields (list[str]): Колонки для проверки соответствия строки
            whereValues (list[str|int|decimal]|_Select): Значения колонок для проверки соответствия строки
            notExecMany (bool): Не применять групповое выполнение для этого запроса
        """
        # если дата окончания переоценки равна None, то значения должны быть другими
        if whereValues[2] is None:
            _Delete.__init__(self, table, whereFields[0:2], whereValues[0:2], notExecMany)
        else:
            _Delete.__init__(self, table, whereFields, whereValues, notExecMany)
        self.cacheKey += 'byDate'

    def prepareStatement(self, context):
        """ Подготовить текст SQL запроса
        Args:
            context (Context): Контекст выполнения запросов
        Returns:
            str: Текст SQL запроса с местами для подстановки значений
        """
        if self.cacheKey in _Delete.cache:
            return _Delete.cache[self.cacheKey]
        tableName = context.tableName(self.table)
        # если дата окончания переоценки равна None, то условие следует изменить
        whereClause = (' `barcode` = ? AND `effectivedate` < ? AND `effectivedateend` %s'
                % ('= ?' if len(self.whereValues) == 3 else 'IS NULL')).replace('?', self.replaceSymbol)
        sql = 'DELETE FROM `%s` WHERE %s' % (tableName, whereClause)
        _Delete.cache[self.cacheKey] = sql
        _Delete.cacheUnite[self.cacheKey] = False
        return sql


class _DeleteOverduePrices(_Delete):
    """ Удаление старых переоценок """

    def __init__(self, table, whereFields, whereValues, notExecMany=False):
        """ Удаление старых переоценок
        Args:
            table (str): Таблица для удаления данных
            whereFields (list[str]): Колонки для проверки соответствия строки
            whereValues (list[str|int|decimal]|_Select): Значения колонок для проверки соответствия строки
            notExecMany (bool): Не применять групповое выполнение для этого запроса
        """
        _Delete.__init__(self, table, whereFields, whereValues, notExecMany)
        self.cacheKey += 'overdue'

    def prepareStatement(self, context):
        """ Подготовить текст SQL запроса
        Args:
            context (Context): Контекст выполнения запросов
        Returns:
            str: Текст SQL запроса с местами для подстановки значений
        """
        if self.cacheKey in _Delete.cache:
            return _Delete.cache[self.cacheKey]
        tableName = context.tableName(self.table)
        sql = 'DELETE FROM `%s` %s AND `effectivedateend` < NOW()' % (tableName, self.formWhereClause(context, self.whereFields, self.whereValues))
        _Delete.cache[self.cacheKey] = sql
        _Delete.cacheUnite[self.cacheKey] = True
        return sql


class _DeleteBeforeDate(_Delete):
    """ MySQL запрос 'Delete' с проверкой, что дата меньше указанной """

    def __init__(self, table, whereFields, whereValues, dateField, dateValue=None, notExecMany=False):
        """ MySQL запрос \"Delete\" с проверкой, что дата меньше указанной
        Args:
            table (str): Таблица для удаления данных
            whereFields (list[str]): Колонки для проверки соответствия строки
            whereValues (list[str|int|decimal]|_Select): Значения колонок для проверки соответствия строки
            dateField (str): Колонка с датой для проверки
            dateValue (str|None): Значение даты для проверки
            notExecMany (bool): Не применять групповое выполнение для этого запроса
        """
        whereValues = self.makeList(whereValues)
        whereFields = self.makeList(whereFields)
        whereFields.append(dateField)
        if dateValue is not None:
            whereValues.append(dateValue)
        self.dateField = dateField
        self.dateValue = dateValue
        _Delete.__init__(self, table, whereFields, whereValues, notExecMany)
        self.cacheKey += 'before'

    def __str__(self):
        return '[DELETE table=%s, whereFields=%s, whereValues=%s, dateField=%s before dateValue=%s]' % (self.table, self.whereFields, self.whereValues, self.dateField, self.dateValue)

    def prepareStatement(self, context):
        """ Подготовить текст SQL запроса
        Args:
            context (Context): Контекст выполнения запросов
        Returns:
            str: Текст SQL запроса с местами для подстановки значений
        """
        if self.cacheKey in _Delete.cache:
            return _Delete.cache[self.cacheKey]
        tableName = context.tableName(self.table)
        whereClause = self.formWhereClause(context, self.whereFields, self.whereValues)
        whereClause = '<'.join(whereClause.rsplit('=', 1))
        sql = 'DELETE FROM `%s` %s' % (tableName, whereClause)
        _Delete.cache[self.cacheKey] = sql
        _Delete.cacheUnite[self.cacheKey] = False
        return sql


class _DeleteAfterDate(_Delete):
    """ MySQL запрос 'Delete' с проверкой, что дата больше указанной """

    def __init__(self, table, whereFields, whereValues, dateField, dateValue=None, notExecMany=False):
        """ MySQL запрос \"Delete\" с проверкой, что дата больше указанной
        Args:
            table (str): Таблица для удаления данных
            whereFields (list[str]): Колонки для проверки соответствия строки
            whereValues (list[str|int|decimal]|_Select): Значения колонок для проверки соответствия строки
            dateField (str): Колонка с датой для проверки
            dateValue (str|None): Значение даты для проверки
            notExecMany (bool): Не применять групповое выполнение для этого запроса
        """
        whereValues = self.makeList(whereValues)
        whereFields = self.makeList(whereFields)
        whereFields.append(dateField)
        if dateValue is not None:
            whereValues.append(dateValue)
        self.dateField = dateField
        self.dateValue = dateValue
        _Delete.__init__(self, table, whereFields, whereValues, notExecMany)
        self.cacheKey += 'after'

    def __str__(self):
        return '[DELETE table=%s, whereFields=%s, whereValues=%s, dateField=%s after dateValue=%s]' % (self.table, self.whereFields, self.whereValues, self.dateField, self.dateValue)

    def prepareStatement(self, context):
        """ Подготовить текст SQL запроса
        Args:
            context (Context): Контекст выполнения запросов
        Returns:
            str: Текст SQL запроса с местами для подстановки значений
        """
        if self.cacheKey in _Delete.cache:
            return _Delete.cache[self.cacheKey]
        tableName = context.tableName(self.table)
        whereClause = self.formWhereClause(context, self.whereFields, self.whereValues)
        whereClause = '>'.join(whereClause.rsplit('=', 1))
        sql = 'DELETE FROM `%s` %s' % (tableName, whereClause)
        _Delete.cache[self.cacheKey] = sql
        _Delete.cacheUnite[self.cacheKey] = False
        return sql


class _DeleteNotNestedItems(_Delete):
    """ Удаление записей с указанным внешним ключём и не входящих в список исключений """

    def __init__(self, table, mainCodeField, mainCodeValue, nestedCodeField, nestedCodeIn):
        pass

    def __str__(self):
        return 'Not execute!'

    def prepareStatement(self, context):
        pass

    def execute(self, context):
        pass


class _IncreaseOrInsertValue(Query):
    """ Увеличение значения поля на заданное значение, если строка существует или добавление новой строки в таблицу """

    def __init__(self, table, increaseField, increaseValue, values, whereFields, whereValues):
        """ Увеличение значения поля на заданное значение, если строка существует или добавление новой строки в таблицу
        Args:
            table (str): Таблица для удаления данных
            increaseField (str): название поля, которое необходимо увеличить
            increaseValue (int|decimal): заданное значение, на которое необходимо увеличить значение поля
            values (dict[str, str|int|decimal]): список пар(название поля - значение), которые необходимо добавить в БД
            whereFields (list[str]): список полей, по которым осуществляется поиск поля для увеличения значения
            whereValues (list[str|int|decimal]): список значений, по которым осуществляется поиск поля для увеличения значения
        """
        self.table = table
        self.increaseField = increaseField
        self.increaseValue = increaseValue
        self.values = values
        self.whereFields = self.makeList(whereFields)
        self.whereValues = self.makeList(whereValues)

    def __str__(self):
        return '[INCREASE VALUE table=%s, field=%s, value=%s, whereFields=%s, whereValues=%s]' % (self.table, self.increaseField, self.increaseValue, self.whereFields, self.whereValues)

    def execute(self, context):
        """ Выполнить SQL запрос
        Args:
            context (Context): Контекст выполнения запросов
        """
        # Поиск поля для увеличения значения
        result = _Select(self.table, [self.increaseField], self.whereFields, self.whereValues).execute(context)

        if result is not None and len(result) > 0:
            self.increaseValue += result[0][0]
        self.values[self.increaseField] = self.increaseValue
        self.values.update(list(zip(self.whereFields, self.whereValues)))
        _Insert(self.table, self.values, onDuplicateKeyUpdate=True).execute(context)
